"""
Going to Camp Web Searching Utilities
"""

import json
import logging
import sys
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple, Union

import requests
from fake_useragent import UserAgent
from pydantic import ValidationError

from camply.containers import AvailableResource, CampgroundFacility, RecreationArea
from camply.containers.base_container import GoingToCampEquipment
from camply.containers.gtc_api_responses import ResourceLocation
from camply.providers.base_provider import BaseProvider, ProviderSearchError
from camply.providers.going_to_camp.rec_areas import RECREATION_AREAS
from camply.utils import make_list
from camply.utils.logging_utils import log_sorted_response

logger = logging.getLogger(__name__)

NON_GROUP_EQUIPMENT = -32768

CAMP_SITE = -2147483648
OVERFLOW_SITE = -2147483647
GROUP_SITE = -2147483643

ENDPOINTS = {
    "CAMP_DETAILS": "https://{}/api/maps",
    "DAILY_AVAILABILITY": "https://{}/api/availability/resourcedailyavailability",
    "LIST_CAMPGROUNDS": "https://{}/api/resourceLocation",
    "LIST_EQUIPMENT": "https://{}/api/equipment",
    "LIST_RESOURCE_CATEGORY": "https://{}/api/resourcecategory",
    "LIST_RESOURCE_STATUS": "https://{}/api/availability/resourcestatus",
    "MAPDATA": "https://{}/api/availability/map",
    "SITE_DETAILS": "https://{}/api/resource/details",
    "ATTRIBUTE_DETAILS": "https://{}/api/attribute/filterable",
}


class GoingToCamp(BaseProvider):
    """
    Going To Camp API provider
    """

    @classmethod
    def find_recreation_areas(
        cls, search_string: Optional[str] = None, **kwargs
    ) -> List[RecreationArea]:
        """
        Find Matching Recreation Areas based on search string

        Parameters
        ----------
        search_string: Optional[str]
            Search Keyword(s)

        Returns
        -------
        filtered_responses: List[RecreationArea]
            Array of Matching Recreation Areas
        """
        if search_string is not None:
            logger.info(f'Searching for Recreation Areas matching: "{search_string}"')

        if not search_string or search_string == "":
            rec_areas = RECREATION_AREAS.values()
            log_sorted_response(rec_areas)
            return rec_areas

        rec_areas = []
        for _, rec_area in RECREATION_AREAS.items():
            if (
                search_string.lower() in rec_area.recreation_area.lower()
                or search_string.lower() in rec_area.recreation_area_location.lower()
            ):
                rec_areas.append(rec_area)

        log_sorted_response(rec_areas)

        return rec_areas

    def rec_area_lookup(self, rec_area_id: int) -> Tuple[str, RecreationArea]:
        """
        Lookup a recreation area by ID

        Parameters
        ----------
        rec_area_id: int
            The recreation area ID to lookup

        Returns
        -------
        domain_name, rec_ara: Tuple[str, RecreationArea]
            The rec area's domain name and the recreation area object
        """
        for domain_name, rec_area in RECREATION_AREAS.items():
            if str(rec_area.recreation_area_id) == str(rec_area_id):
                return domain_name, rec_area

    def find_campgrounds(
        self,
        search_string: Optional[str] = None,
        rec_area_id: Optional[List[int]] = None,
        campground_id: Optional[List[int]] = None,
        campsite_id: Optional[List[int]] = None,
        **kwargs,
    ) -> List[CampgroundFacility]:
        """
        Find Campgrounds Given a Set of Search Criteria

        Parameters
        ----------
        search_string: Optional[str]
            Search Keyword(s)
        rec_area_id: Optional[List[int]]
            Recreation Area ID by which to filter
        campground_id: Optional[List[int]]
            ID of the Campground

        Returns
        -------
        facilities: List[CampgroundFacility]
            Array of Matching Campgrounds
        """
        if rec_area_id in (None, [], ()):
            logger.error(
                "This provider requires --rec-area to be specified when seaching for campsites"
            )
            sys.exit(1)

        return self.find_facilities_per_recreation_area(
            rec_area_id=rec_area_id,
            campground_id=campground_id,
            search_string=search_string,
        )

    def _get_attr_val(self, attribute, attribute_detail) -> any:
        for attr_value in attribute.get("values", []):
            for attribute_enum_detail in attribute_detail.get("values"):
                if attribute_enum_detail["enumValue"] == attr_value:
                    return _fetch_nested_key(
                        attribute_enum_detail, "localizedValues", 0, "displayName"
                    )

    def get_site_details(self, rec_area_id: int, resource_id: int):
        """
        Get the details about a site in a recreation area

        Parameters
        ----------
        rec_area_id: int
            Recreation Area ID by which to filter
        resource_id: int

        Returns
        -------
        details: Dict[str, str]
            The details about the site
        """
        if not hasattr(self, "_attribute_details"):
            self._attribute_details = self._api_request(
                rec_area_id, "ATTRIBUTE_DETAILS"
            )
        attribute_details = self._attribute_details

        site_details = self._api_request(
            rec_area_id, "SITE_DETAILS", {"resourceId": resource_id}
        )
        site_attributes = {}
        for attribute in site_details["definedAttributes"]:
            attribute_detail = attribute_details[
                f"{attribute['attributeDefinitionId']}"
            ]
            attribute_name = _fetch_nested_key(
                attribute_detail, "localizedValues", 0, "displayName"
            )
            attribute_value = attribute.get("value")
            attribute_values = []
            # Attribute a multi-value enum
            if not attribute_value:
                attr_value = self._get_attr_val(attribute, attribute_detail)
                if not attr_value:
                    continue
                attribute_values.append(attr_value)
            else:
                attribute_values.append(f"{attribute_value}")

            site_attributes[attribute_name] = ",".join(attribute_values)
        site_details["site_attributes"] = site_attributes

        return site_details

    def get_reservation_link(
        self,
        rec_area_domain_name,
        resource_location_id,
        map_id,
        equipment_id,
        sub_equipment_id,
        party_size,
        start_date,
        end_date,
    ):
        """
        Generate a URL which a site can be booked

        Returns
        -------
        url: str
            The reservation link URL

        """
        if not sub_equipment_id:
            sub_equipment_id = ""

        return (
            "https://%s/create-booking/results?mapId=%s"
            "&bookingCategoryId=0"
            "&startDate=%s"
            "&endDate=%s"
            "&isReserving=true"
            "&equipmentId=%s"
            "&subEquipmentId=%s"
            "&partySize=%s"
            "&resourceLocationId=%s"
            % (
                rec_area_domain_name,
                map_id,
                start_date.isoformat(),
                end_date.isoformat(),
                equipment_id,
                sub_equipment_id,
                party_size,
                resource_location_id,
            )
        )

    def find_facilities_per_recreation_area(
        self,
        rec_area_id: Optional[Union[List[int], int]] = None,
        campground_id: Optional[Union[List[int], int]] = None,
        search_string: Optional[str] = None,
        **kwargs,
    ) -> List[CampgroundFacility]:
        """
        Find Matching Campsites by Recreation Area

        Parameters
        ----------
        rec_area_id: Optional[Union[List[int], int]]
            Recreation Area ID
        campground_id: Optional[Union[List[int], int]]
            Campground IDs
        search_string: Optional[str]
            A string to search for in the facility name

        Returns
        -------
        campgrounds: List[CampgroundFacility]
            Array of Matching Campsites
        """
        rec_area_id = make_list(rec_area_id, coerce=int)[0]
        logger.info(
            f"Retrieving Facility Information for Recreation Area ID: `{rec_area_id}`."
        )

        rec_area = None
        for _, ra in RECREATION_AREAS.items():
            if str(ra.recreation_area_id) == str(rec_area_id):
                rec_area = ra
        if not rec_area:
            logger.error(f"Recreation area '{rec_area_id}' does not exist.")
            sys.exit(1)

        self.campground_details = {}
        api_response = self._api_request(rec_area_id, "LIST_CAMPGROUNDS")

        filtered_facilities = self._filter_facilities_responses(
            rec_area_id, facilities=api_response
        )

        campgrounds = []
        # Fetch campgrounds details for all facilities
        for camp_details in self._api_request(rec_area_id, "CAMP_DETAILS"):
            self.campground_details[camp_details["resourceLocationId"]] = camp_details

        # If a search string is provided, make sure every facility name contains
        # the search string
        if search_string and search_string not in [[], (), ""]:
            filtered_facilities = [
                f
                for f in filtered_facilities
                if search_string.lower() in f.resource_location_name.lower()
            ]

        for facility in filtered_facilities:
            _, campground_facility = self._process_facilities_responses(
                rec_area, facility=facility
            )
            if not campground_facility:
                continue
            if not campground_id:
                campgrounds.append(campground_facility)
            campground_strings = make_list(campground_id, coerce=str)
            if (
                campground_id
                and str(campground_facility.facility_id) in campground_strings
            ):
                campgrounds.append(campground_facility)
        logger.info(f"{len(campgrounds)} Matching Campgrounds Found")
        log_sorted_response(response_array=campgrounds)
        return campgrounds

    def _hostname_for(self, recreation_area_id: int) -> str:
        for hostname, recreation_area in RECREATION_AREAS.items():
            if str(recreation_area.recreation_area_id) == str(recreation_area_id):
                return hostname
        return None

    def _api_request(
        self,
        rec_area_id: int,
        endpoint_name: str,
        params: Optional[Dict[str, str]] = None,
    ) -> str:
        if params is None:
            params = {}

        hostname = self._hostname_for(rec_area_id)
        endpoint = ENDPOINTS.get(endpoint_name)
        url = None
        if endpoint:
            url = endpoint.format(hostname)
        user_agent = {"User-Agent": UserAgent(browsers=["chrome"]).random}
        response = requests.get(url=url, headers=user_agent, params=params, timeout=30)
        if response.ok is False:
            error_message = f"Receiving bad data from GoingToCamp API: status_code: {response.status_code}: {response.text}"
            logger.error(error_message)
            raise ConnectionError(error_message)

        return json.loads(response.content)

    def _filter_facilities_responses(
        self, rec_area_id: int, facilities=List[Dict[str, Any]]
    ) -> List[ResourceLocation]:
        """
        Filter Facilities to Actual Reservable Campsites

        Parameters
        ----------
        rec_area_id: int
            Recreation Area ID
        facilities: List[Dict[str, Any]]
            List of facilities

        Returns
        -------
        List[ResourceLocation]
        """
        filtered_facilities = []
        for facil in facilities:
            try:
                location_name = _fetch_nested_key(
                    facil, "localizedValues", 0, "fullName"
                )
                park_alerts = _fetch_nested_key(
                    facil, "park_alerts", "en-US", 0, "messageTitle"
                )
                if not park_alerts:
                    park_alerts = _fetch_nested_key(
                        facil, "park_alerts", "en-CA", 0, "messageTitle"
                    )

                region_name = _fetch_nested_key(facil, "region")

                facility = ResourceLocation(
                    id=None,
                    region_name=region_name if region_name else "",
                    park_alerts=park_alerts,
                    rec_area_id=rec_area_id,
                    resource_categories=facil.get("resourceCategoryIds"),
                    resource_location_id=facil.get("resourceLocationId"),
                    resource_location_name=location_name,
                )
            except ValidationError as ve:
                logger.error("That doesn't look like a valid Campground Facility")
                logger.error(facil)
                raise ProviderSearchError(
                    "Invalid Campground Facility Returned"
                ) from ve

            if not facility.resource_categories:
                continue

            # Resource categories from: /api/resourcecategory
            if any(
                [
                    CAMP_SITE in facility.resource_categories,
                    GROUP_SITE in facility.resource_categories,
                    OVERFLOW_SITE in facility.resource_categories,
                ]
            ):
                filtered_facilities.append(facility)

        return filtered_facilities

    def _process_facilities_responses(
        self, rec_area: RecreationArea, facility: ResourceLocation
    ) -> Tuple[dict, Optional[CampgroundFacility]]:
        """
        Process Facilities Responses to be More Usable

        Parameters
        ----------
        facility: dict

        Returns
        -------
        Tuple[dict, CampgroundFacility]
        """
        self.campground_details[facility.resource_location_id]
        facility.id = _fetch_nested_key(
            self.campground_details, facility.resource_location_id, "mapId"
        )
        if facility.region_name:
            formatted_recreation_area = (
                f"{rec_area.recreation_area}, {facility.region_name}"
            )
        else:
            formatted_recreation_area = f"{rec_area.recreation_area}"

        campground_facility = CampgroundFacility(
            facility_name=facility.resource_location_name,
            recreation_area=formatted_recreation_area,
            facility_id=facility.resource_location_id,
            recreation_area_id=facility.rec_area_id,
            map_id=facility.id,
        )
        return facility, campground_facility

    def _find_matching_resources(self, rec_area_id: int, search_filter: Dict[str, any]):
        results = self._api_request(rec_area_id, "MAPDATA", search_filter)
        availability_details = {
            search_filter["mapId"]: results["resourceAvailabilities"]
        }

        return availability_details, list(results["mapLinkAvailabilities"].keys())

    def list_equipment_types(self, rec_area_id: int) -> Dict[str, int]:
        """
        List equipment types available for a recreation area

        Params
        ------
        rec_area_id: int
            The ID of the recreation area

        Returns
        -------
        types: List[GoingToCampEquipment]
            A list of equipment types available to this rec area
        """
        results = self._api_request(rec_area_id, "LIST_EQUIPMENT")

        equipment_types = []
        # Only allow equipment from non-group equipment category (the 0th
        # element in results)
        for sub_category in results[0]["subEquipmentCategories"]:
            equipment_name = _fetch_nested_key(
                sub_category, "localizedValues", 0, "name"
            )
            equipment_id = sub_category["subEquipmentCategoryId"]
            equipment_types.append(
                GoingToCampEquipment(
                    equipment_name=equipment_name, equipment_type_id=equipment_id
                )
            )

        log_sorted_response(response_array=equipment_types)
        return equipment_types

    def list_site_availability(
        self,
        campground: CampgroundFacility,
        start_date: datetime.date,
        end_date: datetime.date,
        equipment_type_id: Optional[str],
    ) -> List[AvailableResource]:
        """
        Retrieve the Availability for all Sites in a Camp Area

        Sites are filtered on the provided date range and compatible
        equipment.

        Returns
        -------
        available_sites: List[AvailableResource]
            The list of available sites
        """
        search_filter = {
            "mapId": campground.map_id,
            "resourceLocationId": campground.facility_id,
            "bookingCategoryId": 0,
            "startDate": start_date.isoformat(),
            "endDate": end_date.isoformat(),
            "isReserving": True,
            "getDailyAvailability": False,
            "partySize": 1,
            "numEquipment": 1,
            "equipmentCategoryId": NON_GROUP_EQUIPMENT,
            "filterData": [],
        }
        if equipment_type_id:
            search_filter["subEquipmentCategoryId"] = equipment_type_id

        resources, additional_resources = self._find_matching_resources(
            campground.recreation_area_id, search_filter
        )

        # Resources are often deeply nested; fetch nested resources
        for map_id in additional_resources:
            search_filter["mapId"] = map_id
            avail, _ = self._find_matching_resources(
                campground.recreation_area_id, search_filter
            )
            resources.update(avail)

        availabilities = []
        for map_id, resource_details in resources.items():
            for resource_id, availability_details in resource_details.items():
                if availability_details[0]["availability"] == 0:
                    ar = AvailableResource(resource_id=resource_id, map_id=map_id)
                    availabilities.append(ar)

        return availabilities


def _fetch_nested_key(obj: Union[dict, list, object], *keys: str) -> Any:
    """
    Fetch nested keys from dictionaries/lists if the keys exist

    Example
    -------
        mydict = {
            'foo': {
                'bar': 'baz'
            }
        }
        val = _fetch_nested_key(mydict, 'foo', 'bar')
        print(f"Value: {val}")
        Prints: Value: baz
    """
    if (
        not isinstance(obj, dict)
        and not isinstance(obj, list)
        and not isinstance(obj, object)
    ):
        raise AttributeError(
            "`obj` must be of type `dict`, `list`, or `object`, but is not"
        )
    if len(keys) == 0:
        raise AttributeError(
            "At least one key must be specified in `keys:`. None were provided"
        )

    _element = obj
    for key in keys:
        try:
            _element = _element[key]
            if not _element:
                _element = getattr(_element, key)
        except (KeyError, TypeError, AttributeError):
            return None

    return _element



"""
Project Configuration for Yellowstone Variables
"""

from collections import OrderedDict
from os.path import abspath, join
from pathlib import Path


class FileConfig:
    """
    File Path Storage Class
    """

    HOME_PATH = abspath(Path.home())
    DOT_CAMPLY_FILE = join(HOME_PATH, ".camply")
    _file_config_file = Path(abspath(__file__))
    _config_dir = _file_config_file.parent

    CAMPLY_DIRECTORY = _config_dir.parent
    ROOT_DIRECTORY = CAMPLY_DIRECTORY.parent

    DOT_CAMPLY_FIELDS = OrderedDict(
        PUSHOVER_PUSH_USER={"default": "", "notes": "Enables Pushover Notifications"},
        PUSHBULLET_API_TOKEN={
            "default": "",
            "notes": "Enables Pushbullet Notifications",
        },
        SLACK_WEBHOOK={"default": "", "notes": "Enables Slack Notifications"},
        TELEGRAM_BOT_TOKEN={"default": "", "notes": "Enables Telegram Notifications"},
        TELEGRAM_CHAT_ID={
            "default": "",
            "notes": "Telegram Notification will be sent here",
        },
        TWILIO_ACCOUNT_SID={"default": "", "notes": "Twilio Account SID"},
        TWILIO_AUTH_TOKEN={"default": "", "notes": "Twilio Auth Token"},
        TWILIO_SOURCE_NUMBER={
            "default": "",
            "notes": "Twilio Source number. E.164 format",
        },
        TWILIO_DEST_NUMBERS={
            "default": "",
            "notes": "Comma-separated list of phone numbers.",
        },
        EMAIL_TO_ADDRESS={
            "default": "",
            "notes": "Email Notifications will be sent here",
        },
        EMAIL_USERNAME={"default": "", "notes": "Email Authorization Login Username"},
        EMAIL_PASSWORD={"default": "", "notes": "Email Authorization Login Password"},
        EMAIL_SMTP_SERVER={
            "default": "smtp.gmail.com",
            "notes": "Email Authorization SMTP Server Address",
        },
        EMAIL_SMTP_PORT={
            "default": 465,
            "notes": "Email Authorization SMTP Server Port",
        },
        EMAIL_FROM_ADDRESS={
            "default": "camply@juftin.com",
            "notes": "Email Notifications Will Come From this Email",
        },
        EMAIL_SUBJECT_LINE={
            "default": "Camply Notification",
            "notes": "Email Notifications Will Have This Subject Line",
        },
        PUSHOVER_PUSH_TOKEN={
            "default": "",
            "notes": "Pushover Notifications From Your Custom App " "(not required)",
        },
        NTFY_TOPIC={
            "default": "",
            "notes": "NTFY Notification Topic",
        },
        APPRISE_URL={"default": "", "notes": "Apprise notification URL"},
        WEBHOOK_URL={"default": "", "notes": "Webhook URL"},
        WEBHOOK_HEADERS={"default": "", "notes": "Webhook JSON Headers"},
        RIDB_API_KEY={
            "default": "",
            "notes": "Personal Recreation.gov API Key (not required)",
        },
    )

    PROVIDERS_DIRECTORY = CAMPLY_DIRECTORY.joinpath("providers")
    RESERVE_CALIFORNIA_PROVIDER = PROVIDERS_DIRECTORY.joinpath("reserve_california")
    USEDIRECT_PROVIDER = PROVIDERS_DIRECTORY.joinpath("usedirect")



"""
Going To Camp API search utilities
"""

import logging
import sys
from datetime import datetime, time
from typing import Any, List, Optional, Union

from camply.containers import (
    AvailableCampsite,
    CampgroundFacility,
    RecreationArea,
    SearchWindow,
)
from camply.providers import GoingToCamp
from camply.providers.going_to_camp.going_to_camp_provider import NON_GROUP_EQUIPMENT
from camply.search.base_search import BaseCampingSearch
from camply.utils import make_list
from camply.utils.general_utils import is_list_like

logger = logging.getLogger(__name__)


class SearchGoingToCamp(BaseCampingSearch):
    """
    Searches on GoingToCamp.com for Campsites
    """

    provider_class = GoingToCamp
    list_campsites_supported: bool = False

    @classmethod
    def find_recreation_areas(
        cls, search_string: str, **kwargs
    ) -> List[RecreationArea]:
        """
        Return the GoingToCamp Recreation Areas
        """
        return GoingToCamp.find_recreation_areas(search_string=search_string)

    # noinspection PyUnusedLocal
    def __init__(
        self,
        search_window: Union[SearchWindow, List[SearchWindow]],
        recreation_area: List[int],
        campsites: Optional[Union[List[int], int]] = None,
        weekends_only: bool = False,
        campgrounds: Optional[Union[List[str], str]] = None,
        equipment_id: Optional[int] = None,
        nights: int = 1,
        **kwargs,
    ) -> None:
        """
        Initialize with Search Parameters

        Parameters
        ----------
        search_window: Union[SearchWindow, List[SearchWindow]]
            Search Window tuple containing start date and End Date
        recreation_area: List[int]
            The ID of the recreation area to be searched. This type of this is a
            List[int], but only for compatability with the BaseCampingSeach API.
            This list is validated to contain a single recreation area.
        weekends_only: bool
            Whether to only search for Camping availabilities on the weekends (Friday /
            Saturday nights)
        campgrounds: Union[List[int], int]
            Campground ID or List of Campground IDs
        nights: int
            minimum number of consecutive nights to search per campsite,defaults to 1
        """
        self.provider = GoingToCamp
        super().__init__(
            search_window=search_window,
            weekends_only=weekends_only,
            nights=nights,
            **kwargs,
        )

        self._recreation_area_id = self._validate_rec_area(recreation_area)
        self._campgrounds = campgrounds
        self.weekends_only = weekends_only
        assert (
            any(
                [
                    campsites not in [[], None],
                    campgrounds not in [[], None],
                    recreation_area is not None,
                ]
            )
            is True
        )
        self.campsites = make_list(campsites)
        self.equipment_id = self._validate_equipment(
            equipment_id, self._recreation_area_id
        )
        self.campgrounds = self._get_searchable_campgrounds()

    @classmethod
    def _validate_rec_area(cls, recreation_area: int) -> int:
        if recreation_area in [(), [], None]:
            logger.error("At least one --rec-area must be provided")
            sys.exit(1)
        if is_list_like(recreation_area) is True and len(recreation_area) > 1:
            logger.error(
                "Going To Camp only allows a single recreation area to be searched at a time"
            )
            sys.exit(1)
        return int(make_list(recreation_area)[0])

    @classmethod
    def _validate_equipment(cls, equipment_id: Optional[int], rec_area: int):
        if not equipment_id:
            return

        try:
            return int(equipment_id)
        except ValueError:
            logger.error(
                "Invalid equipment ID. Use the follwoing to get list of "
                "equipment types: "
                f"`camply equipment-types --provider goingtocamp --rec-area {rec_area}`"
            )
            sys.exit(1)

    def get_all_campsites(self) -> List[AvailableCampsite]:
        """
        Search for all campsites matching search criteria.

        Because Going To Camp has no efficient way of filtering
        campsites for multiple campgrounds, this function is limited to listening
        all campsites _within_ a cammpground. Iterating through all sites for
        all campgrounds would likely lead to abuse complaints.

        Returns
        -------
        List[AvailableCampsite]
        """
        available_sites = []
        for search_window in self.search_window:
            current_start_date = search_window.get_current_start_date()
            for campground in self.campgrounds:
                sites = self.campsite_finder.list_site_availability(
                    campground=campground,
                    start_date=current_start_date,
                    end_date=search_window.end_date,
                    equipment_type_id=self.equipment_id,
                )
                for site in sites:
                    site_details = self.campsite_finder.get_site_details(
                        self._recreation_area_id, site.resource_id
                    )
                    nights = (search_window.end_date - current_start_date).days
                    start_dt = datetime.combine(current_start_date, time.min)
                    end_dt = datetime.combine(search_window.end_date, time.min)
                    (
                        rec_area_domain_name,
                        rec_area,
                    ) = self.campsite_finder.rec_area_lookup(
                        rec_area_id=self._recreation_area_id
                    )
                    booking_url = self.campsite_finder.get_reservation_link(
                        rec_area_domain_name,
                        resource_location_id=campground.facility_id,
                        map_id=site.map_id,
                        equipment_id=NON_GROUP_EQUIPMENT,
                        sub_equipment_id=self.equipment_id,
                        party_size=1,
                        start_date=current_start_date,
                        end_date=search_window.end_date,
                    )

                    # Some rec areas have zero-capacity sites, which should not
                    # be viable for camping. Skip all zero-capacity sites.
                    if (
                        not site_details["minCapacity"]
                        or not site_details["maxCapacity"]
                    ):
                        continue

                    available_sites.append(
                        AvailableCampsite(
                            campsite_id=site_details["resourceId"],
                            campsite_site_name=site_details["localizedValues"][0][
                                "name"
                            ],
                            booking_date=start_dt,
                            booking_end_date=end_dt,
                            booking_nights=nights,
                            campsite_loop_name="Unknown",
                            campsite_type=site_details["site_attributes"].get(
                                "Service Type", "Unknown"
                            ),
                            campsite_occupancy=(
                                site_details["minCapacity"],
                                site_details["maxCapacity"],
                            ),
                            campsite_use_type="N/A",
                            availability_status="Available",
                            recreation_area=rec_area.recreation_area,
                            recreation_area_id=self._recreation_area_id,
                            facility_name=campground.facility_name,
                            facility_id=campground.facility_id,
                            booking_url=booking_url,
                        )
                    )

        return available_sites

    def _get_searchable_campgrounds(self) -> List[CampgroundFacility]:
        """
        Return a List of Campgrounds to search

        Returns
        -------
        searchable_campgrounds: List[CampgroundFacility]
            List of searchable campgrounds
        """
        if self._campgrounds in [(), [], None] or not self._recreation_area_id:
            logger.error(
                "You must provide a Campground and Recreation Area to"
                " search campsites with this provider"
            )
            sys.exit(1)

        if self.campsites not in [(), [], None]:
            self.campsites = [int(campsite_id) for campsite_id in self.campsites]
            return self._get_campgrounds_by_campsite_id()

        return self._get_campgrounds_by_recreation_area_id()

    def _get_campgrounds_by_recreation_area_id(self) -> List[CampgroundFacility]:
        """
        Return a List of Campgrounds to search when provided with Recreation Area IDs

        Returns
        -------
        campgrounds: List[CampgroundFacility]
        """
        if self._campgrounds not in [(), [], None]:
            return self.campsite_finder.find_campgrounds(
                rec_area_id=self._recreation_area_id, campground_id=self._campgrounds
            )

        return self.campsite_finder.find_campgrounds(
            rec_area_id=self._recreation_area_id,
        )

    def list_campsite_units(self) -> Any:
        """
        List Campsite Units

        Returns
        -------
        Any
        """
        raise NotImplementedError



"""
CLI Testing: `camply recreation-areas ...`
"""

import logging

from tests.conftest import CamplyRunner, cli_status_checker, vcr_cassette

logger = logging.getLogger(__name__)


@vcr_cassette
def test_search_recreation_areas(cli_runner: CamplyRunner) -> None:
    """
    Search By a Query String with Rec Areas
    """
    test_command = """
    camply \
        recreation-areas \
        --search "Yosemite National Park"
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "2991" in result.output
    cli_status_checker(result=result)



"""
Generic Webhook Notifications
"""

import logging
from typing import List

import requests

from camply.config.notification_config import WebhookConfig
from camply.containers import AvailableCampsite
from camply.containers.data_containers import WebhookBody
from camply.notifications.base_notifications import BaseNotifications

logger = logging.getLogger(__name__)


class WebhookNotifications(BaseNotifications):
    """
    Push Notifications via Webhooks
    """

    last_gasp: bool = False

    def __init__(self):
        super().__init__()
        self.webhook_url = WebhookConfig.WEBHOOK_URL
        self.webhook_headers = WebhookConfig.DEFAULT_HEADERS
        self.webhook_headers.update(WebhookConfig.WEBHOOK_HEADERS)
        self.session.headers = self.webhook_headers
        if self.webhook_url is None:
            warning_message = (
                "Webhook notifications are not configured properly. "
                "To send webhook messages "
                "make sure to run `camply configure` or set the "
                "proper environment variables: "
                "`WEBHOOK_URL` / `WEBHOOK_HEADERS`."
            )
            logger.error(warning_message)
            raise EnvironmentError(warning_message)

    def send_message(self, message: str, **kwargs) -> None:
        """
        Webhooks only send campsite objects, not messages.
        """
        pass

    def send_campsites(self, campsites: List[AvailableCampsite], **kwargs) -> None:
        """
        Send a message with a campsite object

        Parameters
        ----------
        campsites: List[AvailableCampsite]
        """
        webhook_body = WebhookBody(campsites=campsites).json().encode("utf-8")
        response = self.session.post(url=self.webhook_url, data=webhook_body)
        try:
            response.raise_for_status()
        except requests.HTTPError as he:
            logger.warning(
                f"Notifications weren't able to be sent to {self.webhook_url}. "
                "Your configuration might be incorrect."
            )
            raise ConnectionError(response.text) from he



"""
Yellowstone Testing Provider
"""

import logging
from datetime import datetime

import pytest

from camply.containers import AvailableCampsite, SearchWindow
from camply.search import SearchYellowstone
from tests.conftest import vcr_cassette

logger = logging.getLogger(__name__)


@pytest.fixture
def search_window() -> SearchWindow:
    """
    Get A Search Window For September 2023

    Returns
    -------
    SearchWindow
    """
    search_window = SearchWindow(
        start_date=datetime(2023, 9, 1),
        end_date=datetime(2023, 10, 1),
    )
    return search_window


@pytest.fixture
def yellowstone_finder(search_window) -> SearchYellowstone:
    """
    Assemble The Searching Class

    Returns
    -------
    SearchYellowstone
    """
    yellowstone_finder = SearchYellowstone(search_window=search_window)
    logger.info("Yellowstone Campsite Searcher Established.")
    logger.info(f"Search Months: {yellowstone_finder.search_months}")
    return yellowstone_finder


@vcr_cassette
def test_yellowstone_get_all_campsites(yellowstone_finder) -> None:
    """
    Get all of the Yellowstone Campsites
    """
    logger.info("Searching for Campsites")
    all_campsites = yellowstone_finder.get_all_campsites()
    for camp in all_campsites:
        assert isinstance(camp, AvailableCampsite)



"""
GoingToCamp Testing Provider
"""

import logging
from datetime import datetime

import pytest

from camply.containers import AvailableCampsite, SearchWindow
from camply.search import SearchGoingToCamp
from tests.conftest import vcr_cassette

logger = logging.getLogger(__name__)


@pytest.fixture
def search_window() -> SearchWindow:
    """
    Get A Search Window For September 2023

    Returns
    -------
    SearchWindow
    """
    search_window = SearchWindow(
        start_date=datetime(2023, 9, 1),
        end_date=datetime(2023, 9, 2),
    )
    return search_window


@pytest.fixture
def going_to_camp_finder(search_window) -> SearchGoingToCamp:
    """
    Assemble The Searching Class

    Returns
    -------
    SearchGoingToCamp
    """
    gtc_finder = SearchGoingToCamp(
        search_window=search_window,
        recreation_area=[1],  # Long Point Region, Ontario
        campgrounds="-2147483643",  # Waterford North Conservation Area
    )
    logger.info("GoingToCamp Campsite Searcher Established.")
    logger.info(f"Search Months: {gtc_finder.search_months}")
    return gtc_finder


@vcr_cassette
def test_going_to_camp_get_all_campsites(
    going_to_camp_finder: SearchGoingToCamp,
) -> None:
    """
    Get all GTC sites for the rec area
    """
    logger.info("Searching for Campsites")
    all_campsites = going_to_camp_finder.get_all_campsites()
    assert all_campsites
    for camp in all_campsites:
        assert isinstance(camp, AvailableCampsite)



"""
CLI Testing: `camply campgrounds ...`
"""

import logging

from camply import __version__
from tests.conftest import CamplyRunner, cli_status_checker, vcr_cassette

logger = logging.getLogger(__name__)


@vcr_cassette
def test_list_recdotgov_campgrounds(cli_runner: CamplyRunner) -> None:
    """
    Test the Campground Search for a Fire Tower
    """
    test_command = """
    camply \
        campgrounds \
        --provider RecreationDotGov \
        --search "Fire Tower" \
        --state CA
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Furnace Creek Campground" in result.output
    cli_status_checker(result=result)


@vcr_cassette
def test_campground_search_by_recarea_id(cli_runner: CamplyRunner) -> None:
    """
    Search for Campgrounds by RecArea ID
    """
    test_command = """
    camply \
        campgrounds \
        --rec-area 2991
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Bridalveil Creek Campground" in result.output
    cli_status_checker(result=result)


@vcr_cassette
def test_campground_search_by_campsite_id(cli_runner: CamplyRunner) -> None:
    """
    Search for Campgrounds by Campsite ID
    """
    test_command = """
    camply \
        campgrounds \
        --campsite 40107
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Ledgefork" in result.output
    cli_status_checker(result=result)


@vcr_cassette
def test_campground_search_yellowstone(cli_runner: CamplyRunner) -> None:
    """
    Search for Campgrounds Yellowstone Provider
    """
    test_command = """
    camply \
        --debug \
        campgrounds \
        --provider Yellowstone
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert __version__ in result.output
    cli_status_checker(result=result)
    assert "Madison Campground" in result.output
    cli_status_checker(result=result)



"""
CLI Testing - GoingToCamp - Campsites: `camply campsites ... --provider GoingToCamp`
"""

import logging

from tests.conftest import CamplyRunner, cli_status_checker, vcr_cassette

logger = logging.getLogger(__name__)


@vcr_cassette
def test_goingtocamp_list_recareas(cli_runner: CamplyRunner) -> None:
    """
    Testing GoingToCamp - List All RecAreas
    """
    test_command = """
    camply \
      --provider GoingToCamp \
      recreation-areas
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Long Point Region" in result.output
    cli_status_checker(result=result)


@vcr_cassette
def test_goingtocamp_equipment_types(cli_runner: CamplyRunner) -> None:
    """
    Testing GoingToCamp - List Equipment Types
    """
    test_command = """
    camply \
      equipment-types \
      --rec-area 1 \
      --provider goingtocamp
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Trailer up to 25ft" in result.output
    assert "Trailer up to 40ft" in result.output
    cli_status_checker(result=result)


@vcr_cassette
def test_goingtocamp_search_by_equipment_types(cli_runner: CamplyRunner) -> None:
    """
    Testing GoingToCamp - Search By Equipment Types
    """
    test_command = """
    camply \
      campsites \
      --provider goingtocamp \
      --rec-area 1 \
      --start-date 2023-09-01 \
      --end-date 2023-09-02 \
      --equipment-id -32768 \
      --campground -2147483643
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Long Point Region" in result.output
    assert "Waterford North Conservation Area" in result.output
    assert "Reservable Campsites Matching Search Preferences" in result.output
    cli_status_checker(result=result)


@vcr_cassette
def test_goingtocamp_search_nova_scotia(cli_runner: CamplyRunner) -> None:
    """
    Testing GoingToCamp - Nova Scotia
    """
    test_command = """
    camply campsites \
        --campground -2147483629 \
        --start-date 2023-07-01 \
        --end-date 2023-07-14 \
        --provider GoingToCamp \
        --rec-area 13
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Whycocomagh Provincial Park" in result.output
    assert "Reservable Campsites Matching Search Preferences" in result.output
    cli_status_checker(result=result)


@vcr_cassette
def test_goingtocamp_search_nova_scotia_yaml(cli_runner: CamplyRunner) -> None:
    """
    Testing GoingToCamp - Nova Scotia - YAML
    """
    test_command = """
    camply campsites \
        --yaml-config tests/yaml/goingtocamp.yaml
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Boylston Provincial Park" in result.output
    assert "Reservable Campsites Matching Search Preferences" in result.output
    cli_status_checker(result=result)


@vcr_cassette
def test_goingtocamp_search_parks_canada_campgrounds(cli_runner: CamplyRunner) -> None:
    """
    Testing GoingToCamp - Parks Canada Campgrounds
    """
    test_command = """
    camply campgrounds --provider GoingToCamp --rec-area 14 --search "Two Jack"
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Banff" in result.output
    assert "Two Jack Lakeside" in result.output
    cli_status_checker(result=result)


@vcr_cassette
def test_goingtocamp_search_parks_canada_campsites(cli_runner: CamplyRunner) -> None:
    """
    Testing GoingToCamp - Parks Canada Campsites
    """
    test_command = """
    camply campsites \
        --provider GoingToCamp \
        --rec-area 14 \
        --start-date 2023-07-01 \
        --end-date 2023-07-14 \
        --campground -2147483643
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Banff" in result.output
    assert "Two Jack Lakeside" in result.output
    assert "Reservable Campsites Matching Search Preferences" in result.output
    cli_status_checker(result=result)


@vcr_cassette
def test_goingtocamp_search_manitoba(cli_runner: CamplyRunner) -> None:
    """
    Testing GoingToCamp - Manitoba Campsites
    """
    test_command = """
    camply campsites \
        --provider GoingToCamp \
        --rec-area 15 \
        --start-date 2023-08-01 \
        --end-date 2023-08-14 \
        --campground -2147483632
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Manitoba" in result.output
    assert "Winnipeg Beach Campground" in result.output
    assert "Reservable Campsites Matching Search Preferences" in result.output
    cli_status_checker(result=result)



"""
CLI Testing - RecreationDotGov Tours and Timed Entry
"""

import logging

from tests.conftest import CamplyRunner, cli_status_checker, vcr_cassette

logger = logging.getLogger(__name__)


@vcr_cassette
def test_recdotgov_daily_ticket_campgrounds(cli_runner: CamplyRunner) -> None:
    """
    Provider: RecreationDotGovDailyTicket - Campgrounds
    """
    test_command = """
    camply campgrounds \
        --provider RecreationDotGovDailyTicket \
        --state CO
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Canyons Of The Ancients Museum Tours" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_recdotgov_daily_ticket_campsites(cli_runner: CamplyRunner) -> None:
    """
    Provider: RecreationDotGovDailyTicket - Campsites
    """
    test_command = """
    camply campsites \
        --provider RecreationDotGovDailyTicket \
        --campground 234787 \
        --start-date 2023-09-09 \
        --end-date 2023-09-17
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "total sites found in month of" in result.output
    assert " Chimney Rock National Monument" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_recdotgov_ticket_campgrounds(cli_runner: CamplyRunner) -> None:
    """
    Provider: RecreationDotGovTicket - Campgrounds
    """
    test_command = """
    camply campgrounds --state HI --provider RecreationDotGovTicket
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Haleakala National Park Summit Sunrise" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_recdotgov_ticket_campsites(cli_runner: CamplyRunner) -> None:
    """
    Provider: RecreationDotGovTicket - Campsites
    """
    test_command = """
    camply campsites \
      --provider RecreationDotGovTicket \
      --start-date 2023-06-09 \
      --end-date 2023-06-10 \
      --campground 253731
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Haleakala National Park Summit Sunrise" in result.output
    assert "total sites found in month of" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_recdotgov_timed_entry_campgrounds(cli_runner: CamplyRunner) -> None:
    """
    Provider: RecreationDotGovTimedEntry - Campgrounds
    """
    test_command = """
    camply campgrounds --provider RecreationDotGovTimedEntry --state OR
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Lava River Cave" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_recdotgov_timed_entry_campsites(cli_runner: CamplyRunner) -> None:
    """
    Provider: RecreationDotGovTimedEntry - Campsites
    """
    test_command = """
    camply campsites \
      --provider RecreationDotGovTimedEntry \
      --start-date 2023-06-09 \
      --end-date 2023-06-10 \
      --campground 10089508
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Lava River Cave" in result.output
    assert "total sites found in month of" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_recdotgov_daily_ticket_campsites_equipment(cli_runner: CamplyRunner) -> None:
    """
    Provider: RecreationDotGovDailyTicket - Campsites + Equipment
    """
    test_command = """
    camply campsites \
      --provider RecreationDotGovDailyTicket \
      --start-date 2023-07-06 \
      --end-date 2023-07-07 \
      --campground 300004 \
      --equipment 1300 4
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Filtering Campsites based on Equipment: 1300" in result.output
    assert "Joshua Tree National Park Tours" in result.output
    assert "total sites found in month of" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_list_campsites_ticket(cli_runner: CamplyRunner) -> None:
    """
    Provider: RecreationDotGovTicket - List Campsites
    """
    test_command = """
    camply list-campsites \
        --provider RecreationDotGovTicket \
        --campground 253731
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Searching for campsites to list" in result.output
    assert "Haleakala Sunrise - Summit - (#255)" in result.output
    cli_status_checker(result=result, exit_code_zero=True)


@vcr_cassette
def test_list_campsites_time_entry(cli_runner: CamplyRunner) -> None:
    """
    Provider: RecreationDotGovTimedEntry - List Campsites
    """
    test_command = """
    camply list-campsites \
        --provider RecreationDotGovTimedEntry \
        --campground 10089508
    """
    result = cli_runner.run_camply_command(command=test_command)
    assert "Searching for campsites to list" in result.output
    assert "Lava River Cave Vehicle Reservations - (#10089509)" in result.output
    cli_status_checker(result=result, exit_code_zero=True)



"""
camply search __init__ file
"""

from typing import Dict, List, Type

from camply.search.base_search import BaseCampingSearch
from camply.search.search_going_to_camp import SearchGoingToCamp
from camply.search.search_recreationdotgov import (
    SearchRecreationDotGov,
    SearchRecreationDotGovDailyTicket,
    SearchRecreationDotGovDailyTimedEntry,
    SearchRecreationDotGovTicket,
    SearchRecreationDotGovTimedEntry,
)
from camply.search.search_usedirect import (
    SearchAlabamaStateParks,
    SearchArizonaStateParks,
    SearchFairfaxCountyParks,
    SearchFloridaStateParks,
    SearchMaricopaCountyParks,
    SearchMinnesotaStateParks,
    SearchMissouriStateParks,
    SearchNorthernTerritory,
    SearchOhioStateParks,
    SearchOregonMetro,
    SearchReserveCalifornia,
    SearchVirginiaStateParks,
)
from camply.search.search_yellowstone import SearchYellowstone

# Register Providers Here with their Search class
__search_providers__: List[Type[BaseCampingSearch]] = [
    SearchRecreationDotGov,
    SearchYellowstone,
    SearchGoingToCamp,
    # UseDirect
    SearchReserveCalifornia,
    SearchAlabamaStateParks,
    SearchArizonaStateParks,
    SearchFloridaStateParks,
    SearchMinnesotaStateParks,
    SearchMissouriStateParks,
    SearchOhioStateParks,
    SearchVirginiaStateParks,
    SearchNorthernTerritory,
    SearchFairfaxCountyParks,
    SearchMaricopaCountyParks,
    SearchOregonMetro,
    # Tours and Timed Entry (RecDotGov)
    SearchRecreationDotGovTicket,
    SearchRecreationDotGovTimedEntry,
    SearchRecreationDotGovDailyTicket,
    SearchRecreationDotGovDailyTimedEntry,
]

CAMPSITE_SEARCH_PROVIDER: Dict[str, Type[BaseCampingSearch]] = {
    provider.provider_class.__name__: provider for provider in __search_providers__
}



"""
Campsite Searching: Recreation.gov
"""

import json
import logging
from datetime import datetime, timedelta
from itertools import chain
from typing import Any, Dict, List, Optional, Tuple, Union

import pandas as pd
import requests

from camply.config import RecreationBookingConfig, RIDBConfig
from camply.containers import AvailableCampsite
from camply.containers.api_responses import (
    CampsiteAvailabilityResponse,
    CampsiteResponse,
    RecDotGovCampsite,
    RecDotGovCampsiteResponse,
)
from camply.containers.data_containers import CampsiteLocation
from camply.providers.recreation_dot_gov.recdotgov_provider import RecreationDotGovBase
from camply.utils import api_utils

logger = logging.getLogger(__name__)


class RecreationDotGov(RecreationDotGovBase):
    """
    Recreation.gov: Campsite Searcher
    """

    facility_type = RIDBConfig.CAMPGROUND_FACILITY_FIELD_QUALIFIER
    resource_api_path = RIDBConfig.CAMPSITE_API_PATH
    activity_name = "CAMPING"
    api_response_class = CampsiteResponse
    api_base_path = RecreationBookingConfig.API_BASE_PATH
    api_search_result_class = RecDotGovCampsite
    api_search_result_key = "campsite_id"

    def paginate_recdotgov_campsites(
        self, facility_id: int, equipment: Optional[List[str]] = None
    ) -> List[RecDotGovCampsite]:
        """
        Paginate through the RecDotGov Campsite Metadata
        """
        results = 0
        continue_paginate = True
        endpoint_url = api_utils.generate_url(
            scheme=RecreationBookingConfig.API_SCHEME,
            netloc=RecreationBookingConfig.API_NET_LOC,
            path="api/search/campsites",
        )
        fq_list = [f"asset_id:{facility_id}"]
        if isinstance(equipment, list) and len(equipment) > 0:
            for item in equipment:
                fq_list.append(f"campsite_equipment_name:{item}")
        params = {
            "start": 0,
            "size": 1000,
            "fq": fq_list,
            "include_non_site_specific_campsites": True,
        }
        campsites: List[RecDotGovCampsite] = []
        while continue_paginate is True:
            response = self.make_recdotgov_request_retry(
                method="GET",
                url=endpoint_url,
                params=params,
            )
            returned_data = json.loads(response.content)
            campsite_response = RecDotGovCampsiteResponse(**returned_data)
            campsites += campsite_response.campsites
            results += campsite_response.size
            params.update(start=results)
            if results == campsite_response.total:
                continue_paginate = False
        return campsites

    def make_recdotgov_availability_request(
        self,
        campground_id: int,
        month: datetime,
    ) -> requests.Response:
        """
        Make a request to the RecreationDotGov API

        Parameters
        ----------
        campground_id
        month

        Returns
        -------
        requests.Response
        """
        api_endpoint = self._rec_availability_get_endpoint(
            path=f"{campground_id}/{RecreationBookingConfig.API_MONTH_PATH}"
        )
        formatted_month = month.strftime("%Y-%m-01T00:00:00.000Z")
        query_params = {"start_date": formatted_month}
        return self.make_recdotgov_request(
            method="GET",
            url=api_endpoint,
            params=query_params,
        )

    @classmethod
    def _items_to_unique_dicts(
        cls, item: Union[List[Dict[str, Any]], pd.Series]
    ) -> List[Dict[str, Any]]:
        """
        Ensure the proper items are parsed for equipment and attributes
        """
        if isinstance(item, pd.Series):
            list_of_dicts = list(chain.from_iterable(item.tolist()))
            unique_list_of_dicts = [
                dict(s) for s in {frozenset(d.items()) for d in list_of_dicts}
            ]
            return unique_list_of_dicts
        else:
            return item

    @classmethod
    def _get_equipment_attributes_location(
        cls,
        campsite_id: int,
        campsite_metadata: pd.DataFrame,
    ) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]], Optional[CampsiteLocation]]:
        """
        Index a DataFrame in a Complicated Way
        """
        try:
            equipment = campsite_metadata.at[campsite_id, "permitted_equipment"]
        except LookupError:
            equipment = None
        try:
            attributes = campsite_metadata.at[campsite_id, "attributes"]
        except LookupError:
            attributes = None
        try:
            latitude = campsite_metadata.at[campsite_id, "latitude"]
            longitude = campsite_metadata.at[campsite_id, "longitude"]
            location = CampsiteLocation(
                latitude=latitude,
                longitude=longitude,
            )
        except LookupError:
            location = None
        equipment = cls._items_to_unique_dicts(item=equipment)
        attributes = cls._items_to_unique_dicts(item=attributes)
        return equipment, attributes, location

    @classmethod
    def process_campsite_availability(
        cls,
        availability: Dict[str, Any],
        recreation_area: str,
        recreation_area_id: int,
        facility_name: str,
        facility_id: int,
        month: datetime,
        campsite_metadata: pd.DataFrame,
    ) -> List[Optional[AvailableCampsite]]:
        """
        Parse the JSON Response and return availabilities

        Parameters
        ----------
        availability: Dict[str, Any]
            API Response
        recreation_area: str
            Name of Recreation Area
        recreation_area_id: int
            ID of Recreation Area
        facility_name: str
            Campground Facility Name
        facility_id: int
            Campground Facility ID
        month: datetime
            Month to Process
        campsite_metadata: pd.DataFrame
            Metadata Fetched from the Recreation.gov API about the Campsites

        Returns
        -------
        total_campsite_availability: List[Optional[AvailableCampsite]]
            Any monthly availabilities
        """
        total_campsite_availability: List[Optional[AvailableCampsite]] = []
        campsite_data = CampsiteAvailabilityResponse(**availability)
        for campsite_id, site_related_data in campsite_data.campsites.items():
            for (
                matching_date,
                availability_status,
            ) in site_related_data.availabilities.items():
                if (
                    availability_status
                    not in RecreationBookingConfig.CAMPSITE_UNAVAILABLE_STRINGS
                ):
                    booking_url = (
                        f"{RecreationBookingConfig.CAMPSITE_BOOKING_URL}/{campsite_id}"
                    )
                    (
                        equipment,
                        attributes,
                        location,
                    ) = cls._get_equipment_attributes_location(
                        campsite_id=campsite_id, campsite_metadata=campsite_metadata
                    )
                    available_campsite = AvailableCampsite(
                        campsite_id=campsite_id,
                        booking_date=matching_date,
                        booking_end_date=matching_date + timedelta(days=1),
                        booking_nights=1,
                        campsite_site_name=site_related_data.site,
                        campsite_loop_name=site_related_data.loop,
                        campsite_type=site_related_data.campsite_type,
                        campsite_occupancy=(
                            site_related_data.min_num_people,
                            site_related_data.max_num_people,
                        ),
                        campsite_use_type=site_related_data.type_of_use,
                        availability_status=availability_status,
                        recreation_area=recreation_area,
                        recreation_area_id=recreation_area_id,
                        facility_name=facility_name,
                        facility_id=facility_id,
                        booking_url=booking_url,
                        permitted_equipment=equipment,
                        campsite_attributes=attributes,
                        location=location,
                    )
                    total_campsite_availability.append(available_campsite)
        return total_campsite_availability



"""
Pydantic model for YAML files
"""

import datetime
from enum import Enum
from typing import List, Optional, Tuple, Union

from pydantic import Field, validator

from camply.config import SearchConfig
from camply.containers import CamplyModel
from camply.search import CAMPSITE_SEARCH_PROVIDER

ArrayOrSingleInt = Optional[Union[int, List[int]]]
ArrayOrSingleStr = Optional[Union[str, List[str]]]
ArrayOrSingle = Optional[Union[ArrayOrSingleInt, ArrayOrSingleStr]]
EquipmentTuple = Optional[Tuple[str, int]]
ArrayOrSingleEquipment = Optional[Union[EquipmentTuple, List[EquipmentTuple]]]


class StrEnum(str, Enum):
    """
    String Enum
    """


ProviderEnum = StrEnum(
    "ProviderEnum", {value: value for value in CAMPSITE_SEARCH_PROVIDER.keys()}
)
ProviderEnum.__doc__ = "Campsite Provider Names"


class YamlSearchFile(CamplyModel):
    """
    Campsite Search Data Model
    """

    provider: ProviderEnum = Field(
        description="Campsite provider", default="RecreationDotGov"
    )
    recreation_area: ArrayOrSingle = None
    campgrounds: ArrayOrSingle = None
    campsites: ArrayOrSingle = None
    start_date: Union[datetime.date, List[datetime.date]]
    end_date: Union[datetime.date, List[datetime.date]]
    days: Optional[List[str]] = None
    weekends: bool = False
    nights: int = 1
    continuous: bool = True
    polling_interval: int = SearchConfig.RECOMMENDED_POLLING_INTERVAL
    notifications: ArrayOrSingleStr = "silent"
    search_forever: bool = False
    search_once: bool = False
    notify_first_try: bool = False
    equipment: ArrayOrSingleEquipment = None
    offline_search: bool = False
    offline_search_path: Optional[str] = None

    @validator("provider", pre=True)
    def validate_provider(cls, value):
        """
        Validate provider
        """
        lowercase_enum_dict = {
            key.lower(): key for key in ProviderEnum.__members__.keys()
        }
        if value.lower() in lowercase_enum_dict.keys():
            return lowercase_enum_dict[value.lower()]
        else:
            return value

    @validator("equipment", pre=True)
    def validate_equipment(cls, value) -> ArrayOrSingleEquipment:
        """
        Validate equipment
        """
        equipment_tuple_length = 2
        if (
            isinstance(value, list)
            and len(value) == equipment_tuple_length
            and isinstance(value[0], str)
        ):
            return [tuple(value)]
        else:
            return value



"""
Notification Testing
"""

from camply import AvailableCampsite
from camply.notifications import PushoverNotifications
from tests.conftest import vcr_cassette


@vcr_cassette
def test_pushover_message():
    """
    Send a Pushover Message
    """
    pusher = PushoverNotifications()
    pusher.send_message(message="This is a test message!")


@vcr_cassette
def test_pushover_campsite(available_campsite: AvailableCampsite):
    """
    Send a Pushover Campsite
    """
    pusher = PushoverNotifications()
    pusher.send_campsites(campsites=[available_campsite])



"""
Camply: Common Exceptions
"""


class CamplyError(Exception):
    """
    Base Camply Error
    """


class SearchError(CamplyError):
    """
    Generic Search Error
    """


class CampsiteNotFoundError(SearchError):
    """
    Campsite not found Error
    """



"""
Camply Command Line Interface
"""

import logging
import sys
from dataclasses import dataclass
from datetime import date, timedelta
from typing import (
    Any,
    Container,
    Dict,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)

import click
from rich import traceback
from rich_click import RichCommand, RichGroup, rich_click

from camply import Yellowstone, __application__, __version__
from camply.config import EquipmentOptions, SearchConfig, logging_config
from camply.config.logging_config import set_up_logging
from camply.containers import SearchWindow
from camply.containers.examples import example_campsite
from camply.notifications import CAMPSITE_NOTIFICATIONS, MultiNotifierProvider
from camply.providers import (
    GoingToCamp,
    RecreationDotGov,
)
from camply.search import CAMPSITE_SEARCH_PROVIDER, BaseCampingSearch
from camply.utils import configure_camply, log_camply, make_list, yaml_utils
from camply.utils.general_utils import days_of_the_week_mapping, handle_search_windows
from camply.utils.logging_utils import log_sorted_response

logging.Logger.camply = log_camply
logger = logging.getLogger(__name__)

DEFAULT_CAMPLY_PROVIDER: str = RecreationDotGov.__name__

rich_click.STYLE_OPTION = "bold green"
rich_click.STYLE_SWITCH = "bold blue"
rich_click.STYLE_METAVAR = "bold red"
rich_click.STYLE_HELPTEXT_FIRST_LINE = "bold blue"
rich_click.STYLE_HELPTEXT = ""
rich_click.STYLE_HEADER_TEXT = "bold green"
rich_click.STYLE_OPTION_DEFAULT = "bold yellow"
rich_click.STYLE_OPTION_HELP = ""
rich_click.STYLE_ERRORS_SUGGESTION = "bold red"
rich_click.STYLE_OPTIONS_TABLE_BOX = "SIMPLE_HEAVY"
rich_click.STYLE_COMMANDS_TABLE_BOX = "SIMPLE_HEAVY"
if logging_config.LOG_HANDLER == "python":
    rich_click.COLOR_SYSTEM = None

try:
    from trogon import tui
except ImportError:

    def tui(*args, **kwargs):
        """
        TUI Placeholder - trogon not installed
        """

        def placeholder(app: click.Group):
            """
            Return the group in place
            """
            return app

        return placeholder


@dataclass
class CamplyContext:
    """
    Context Object Passed Around Application
    """

    debug: Optional[bool] = None
    provider: Optional[str] = None


provider_argument = click.option(
    "--provider",
    show_default=False,
    default=None,
    type=click.Choice(CAMPSITE_SEARCH_PROVIDER.keys(), case_sensitive=False),
    help="Camping Search Provider. Defaults to 'RecreationDotGov'",
    metavar="TEXT",
    envvar="CAMPLY_PROVIDER",
)
debug_option = click.option(
    "--debug/--no-debug", default=None, help="Enable extra debugging output"
)


def _set_up_debug(debug: Optional[bool] = None) -> None:
    """
    Set up the Camply Debugging Mode
    """
    if debug is None:
        debug = False
    if debug is True:
        set_up_logging(log_level=logging.DEBUG)
        logger.debug("Setting up camply debugging")
        logger.debug("Camply Version: %s", __version__)
        logger.debug("Python Version: %s", sys.version.split(" ")[0])
        logger.debug("Platform: %s", sys.platform)
    traceback.install(show_locals=debug, suppress=[click, rich_click])


def _preferred_provider(context: CamplyContext, command_provider: Optional[str]) -> str:
    """
    Called to get the preferred subcommands provider.

    It establishes rules for the "preferred" provider. That is, when multiple
    providers are elgigible to serve a command, the one that is most specific is
    chosen. Preference is in the following order:

    1. The provider explicitly provided to a camply subcommand
    2. The provider associated with CamplyContext
    3. The default provider

    The preferred provider is returned
    """
    if command_provider:
        provider = command_provider
    elif command_provider is None and context.provider:
        provider = context.provider
    else:
        provider = DEFAULT_CAMPLY_PROVIDER
    logger.info('Using Camply Provider: "%s"', provider)
    return provider


@tui()
@click.group(cls=RichGroup)
@debug_option
@provider_argument
@click.version_option(version=__version__, prog_name=__application__)
@click.pass_context
def camply_command_line(
    ctx: click.core.Context, debug: bool, provider: Optional[str]
) -> None:
    """
    Welcome to camply, the campsite finder.

    Finding reservations at sold out campgrounds can be
    tough. That's where camply comes in. It searches the APIs of booking services like
    https://recreation.gov (which indexes thousands of campgrounds across the USA) to continuously
    check for cancellations and availabilities to pop up. Once a campsite becomes available,
    camply sends you a notification to book your spot!

    \b

    visit the camply documentation at https://juftin.com/camply
    """
    set_up_logging(log_level=None if debug is False else logging.INFO)
    logger.camply("camply, the campsite finder ⛺️")
    ctx.obj = CamplyContext(debug=debug, provider=provider)
    _set_up_debug(debug=debug)


@camply_command_line.command(cls=RichCommand)
@debug_option
@click.pass_obj
def configure(context: CamplyContext, debug: bool) -> None:
    """
    Set up camply configuration file with an interactive console

    In order to send notifications through camply you must set up some authorization values.
    Whether you need to set up pushover notifications (push notifications on your phone,
    your pushover account can be set up at https://pushover.net) or Email messages, everything
    can be done through the configure command. The end result is a file called .camply in your
    home folder.
    """
    if context.debug is None:
        context.debug = debug
        _set_up_debug(debug=context.debug)
    configure_camply.generate_dot_camply_file()


# Shared Arguments
search_argument = click.option(
    "--search",
    default=None,
    help="Search for Campgrounds or Recreation Areas by search string.",
)
state_argument = click.option("--state", default=None, help="Filter by US state code.")
campsite_id_argument = click.option(
    "--campsite", default=None, multiple=True, help="Add individual Campsites by ID."
)
rec_area_argument = click.option(
    "--rec-area",
    default=None,
    multiple=True,
    help="Add Recreation Areas (comprised of campgrounds) by ID.",
)
campground_argument = click.option(
    "--campground",
    default=None,
    multiple=True,
    help="Add individual Campgrounds by ID.",
)


@camply_command_line.command(cls=RichCommand)
@rec_area_argument
@provider_argument
@click.pass_obj
def equipment_types(
    context: CamplyContext,
    rec_area: Optional[int] = None,
    provider: str = DEFAULT_CAMPLY_PROVIDER,
) -> None:
    """
    Get a list of supported equipment

    This command returns supported equipment for the current provider/recreation area.
    Equipment are camping equipment that can be used at a campsite. Different providers
    and recreation areas have different types of equipment for which reservations can be made.
    """
    provider = _preferred_provider(context, provider)
    if not rec_area and provider == GoingToCamp.__name__:
        logger.error(
            "This provider requires --rec-area to be specified when listing equipment types"
        )
        sys.exit(1)

    if provider == GoingToCamp.__name__:
        GoingToCamp().list_equipment_types(rec_area[0])
    elif provider.startswith(RecreationDotGov.__name__):
        log_sorted_response(response_array=EquipmentOptions.__all_accepted_equipment__)
    else:
        logger.warning(
            "That Provider doesn't support equipment based searching, yet 🙂"
        )
    sys.exit(0)


@camply_command_line.command(cls=RichCommand)
@search_argument
@state_argument
@debug_option
@provider_argument
@click.pass_obj
def recreation_areas(
    context: CamplyContext,
    search: Optional[str],
    state: Optional[str],
    debug: bool,
    provider: str = DEFAULT_CAMPLY_PROVIDER,
) -> None:
    """
    Search for Recreation Areas and list them

    Search for Recreation Areas and their IDs. Recreation Areas are places like
    National Parks and National Forests that can contain one or many campgrounds.
    """
    provider = _preferred_provider(context, provider)
    if context.debug is None:
        context.debug = debug
        _set_up_debug(debug=context.debug)
    if all(
        [
            search is None,
            state is not None,
            provider in [Yellowstone.__name__, GoingToCamp.__name__],
        ]
    ):
        # State Filtering Not Supported
        logger.error(
            f"{provider} does not support filtering recreation areas by state. Leave --state blank."
        )
        sys.exit(1)
    if provider == GoingToCamp.__name__:
        rec_area_finder = GoingToCamp()
    elif provider.startswith(RecreationDotGov.__name__):
        rec_area_finder = RecreationDotGov()
    else:
        rec_area_finder = CAMPSITE_SEARCH_PROVIDER[provider]
    params = {}
    if state is not None:
        params.update({"state": state})
    rec_area_finder.find_recreation_areas(search_string=search, **params)


@camply_command_line.command(cls=RichCommand)
@search_argument
@state_argument
@rec_area_argument
@campground_argument
@campsite_id_argument
@provider_argument
@debug_option
@click.pass_obj
def campgrounds(
    context: CamplyContext,
    debug: bool,
    search: Optional[str] = None,
    state: Optional[str] = None,
    rec_area: Optional[int] = None,
    campground: Optional[int] = None,
    campsite: Optional[int] = None,
    provider: Optional[str] = DEFAULT_CAMPLY_PROVIDER,
) -> None:
    """
    Search for Campgrounds (inside of Recreation Areas) and list them

    Search for Campgrounds and their IDs. Campgrounds are facilities inside of
    Recreation Areas that contain campsites. Most 'campgrounds' are areas made up of
    multiple campsites, others are facilities like fire towers or cabins that might only
    contain a single 'campsite' to book.
    """
    provider = _preferred_provider(context, provider)
    if context.debug is None:
        context.debug = debug
        _set_up_debug(debug=context.debug)
    if all(
        [
            search is None,
            state is None,
            len(rec_area) == 0,
            len(campground) == 0,
            len(campsite) == 0,
            provider not in [Yellowstone.__name__, GoingToCamp.__name__],
        ]
    ):
        logger.error(
            "You must add a --search, --state, --campground, --campsite, "
            "or --rec-area parameter to search for Campgrounds."
        )
        sys.exit(1)
    search_provider_class = CAMPSITE_SEARCH_PROVIDER[provider]
    camp_finder = search_provider_class.provider_class()
    params = {}
    if state is not None:
        params.update({"state": state})
    camp_finder.find_campgrounds(
        search_string=search,
        rec_area_id=make_list(rec_area, coerce=int),
        campground_id=make_list(campground, coerce=int),
        campsite_id=make_list(campsite, coerce=int),
        **params,
    )


# `campsite` arguments
start_date_argument = click.option(
    "--start-date",
    multiple=True,
    default=None,
    help="(YYYY-MM-DD) Start of Search window. You will be arriving this day.",
)
end_date_argument = click.option(
    "--end-date",
    multiple=True,
    default=None,
    help="(YYYY-MM-DD) End of Search window. You will be checking out this day.",
)
weekends_argument = click.option(
    "--weekends",
    is_flag=True,
    show_default=True,
    default=False,
    help="Only search for weekend bookings (Fri/Sat nights).",
)
nights_argument = click.option(
    "--nights",
    show_default=True,
    default=1,
    type=click.INT,
    help="Search for campsite stays with consecutive nights. "
    "Defaults to 1 which returns all campsites found.",
)
continuous_argument = click.option(
    "--continuous",
    is_flag=True,
    show_default=True,
    default=False,
    help="Continuously check for a campsite to become available, and quit once "
    "at least one campsite is found.",
)
polling_interval_argument = click.option(
    "--polling-interval",
    default=None,
    type=click.INT,
    help="Enables continuous searching. How often to wait in between "
    "checks (in minutes). Defaults to 10, cannot be less than 5.",
)

_joined_notifications = [f"`{item}`" for item in CAMPSITE_NOTIFICATIONS.keys()]
notification_kwargs = {
    "multiple": True,
    "show_default": True,
    "metavar": "TEXT",
    "type": click.Choice(
        choices=list(CAMPSITE_NOTIFICATIONS.keys()), case_sensitive=False
    ),
    "default": [],
    "help": "Enables continuous searching. Types of notifications to receive. "
    f"Options available are {', '.join(_joined_notifications)}. "
    "Defaults to `silent` - "
    "which just logs messages to console.",
}
notifications_argument = click.option("--notifications", **notification_kwargs)
notify_first_try_argument = click.option(
    "--notify-first-try",
    is_flag=True,
    flag_value="FLAG-SET",
    help="Enables continuous searching. Whether to send all "
    "non-silent notifications if more than 5 matching "
    "campsites are found on the first try. Defaults to False which "
    "only sends the first 5.",
)
search_forever_argument = click.option(
    "--search-forever",
    is_flag=True,
    flag_value="FLAG-SET",
    help="Enables continuous searching. This method continues to search "
    "after the first availability has been found. The one caveat is "
    "that it will never notify about the same identical campsite for "
    "the same booking date.",
)
search_once_argument = click.option(
    "--search-once",
    is_flag=True,
    default=False,
    help="Enables continuous searching features - but doesn't actually "
    "search continuously. This option is only useful when you want to run camply "
    "periodically in a CRON job fashion but still receive notifications - "
    "it's strongly recommended you enable offline searching as well to "
    "save results between searches.",
)
yaml_config_argument = click.option(
    "--yaml-config",
    "--yml-config",
    default=None,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help="Rather than provide arguments to the command line utility, instead "
    "pass a file path to a YAML configuration file. See the documentation "
    "for more information on how to structure your configuration file.",
)
equipment_argument = click.option(
    "--equipment",
    default=None,
    nargs=2,
    multiple=True,
    help="Search for campsites compatible with your camping equipment. "
    "This argument accepts two options, the equipment name and its length "
    "If you don't want to filter based on length provide a length of 0. Accepted "
    "equipment names include `Tent`, `RV`. `Trailer`, `Vehicle` and are "
    "not case-sensitive.",
)
equipment_id_argument = click.option(
    "--equipment-id",
    default=None,
    help="Search for campsites compatible with specific equipment categories. Going To "
    "Camp uses equipment category IDs for filtering campsites by equipment. Every "
    "recreation area has equipment categories unique to it. "
    "Use `camply equipment-types --provider GoingToCamp --rec-area <rec area id>` "
    "to get a listing of equipment for an area.",
)

offline_search_argument = click.option(
    "--offline-search",
    is_flag=True,
    show_default=True,
    default=False,
    help="When set to True, the campsite search will both save the results of the "
    "campsites it's found, but also load those campsites before beginning a "
    "search for other campsites.",
)
offline_search_path_argument = click.option(
    "--offline-search-path",
    show_default=True,
    default=None,
    type=click.Path(dir_okay=False),
    help="Enables offline search. This is the name of the file to be saved/loaded. "
    "Campsites can be saved as a serialized pickle file or "
    "a JSON file, depending on the file extension. When not specified, "
    "the filename will default to `camply_campsites.json`",
)
day_of_the_week_argument = click.option(
    "--day",
    multiple=True,
    show_default=False,
    type=click.Choice(
        choices=list(days_of_the_week_mapping.keys()), case_sensitive=False
    ),
    metavar="TEXT",
    help="Day(s) of the Week to search.",
)


def _get_equipment(equipment: Optional[List[str]]) -> List[Tuple[str, Optional[int]]]:
    """
    Parse Equipment from CLI Args
    """
    equipment_list = []
    for equipment_name, equipment_length in equipment:
        try:
            equipment_length = round(float(equipment_length), 0)
            if equipment_length == 0:
                equipment_length = None
        except ValueError:
            equipment_length = None
        equipment_list.append((equipment_name, equipment_length))
    return equipment_list


def _validate_campsites(
    rec_area: Optional[int],
    campground: Optional[int],
    campsite: Optional[int],
    start_date: Tuple[str],
    end_date: Tuple[str],
    provider: str,
    yaml_config: Optional[str],
    continuous: bool,
    polling_interval: int,
    notifications: List[str],
    notify_first_try: bool,
    search_forever: bool,
    search_once: bool,
    day: Optional[Tuple[str]],
    **kwargs: Dict[str, Any],
) -> Tuple[bool, List[SearchWindow], Set[int]]:
    """
    Validate the campsites portion of the CLI

    Also, indicate whether the search should be continuous.

    Parameters
    ----------
    rec_area: Optional[int]
    campground: Optional[int]
    campsite: Optional[int]
    start_date: Optional[str]
    end_date: Optional[str]
    provider: str
    yaml_config: Optional[str]
    continuous: bool
    polling_interval: int
    notifications: List[str]
    notify_first_try: bool
    search_forever: bool
    day: Optional[Tuple[str]]
    **kwargs: Dict[str, Any]

    Returns
    -------
    Tuple[bool, List[SearchWindow], Set[int]]
        Tuple containing continuous run eval, search_windows,
        and days of the week
    """
    if provider.startswith(RecreationDotGov.__name__) and all(
        [
            len(rec_area) == 0,
            len(campground) == 0,
            len(campsite) == 0,
            yaml_config is None,
        ]
    ):
        logger.error(
            "To search for Recreation.gov Campsites you must provide "
            "either the --rec-area, --campground, or --campsite "
            "parameters."
        )
        sys.exit(1)
    if yaml_config is None:
        search_windows = handle_search_windows(start_date=start_date, end_date=end_date)
    else:
        search_windows = ()
    days_of_the_week = None
    if day is not None:
        days_of_the_week = {days_of_the_week_mapping[item] for item in day}
    if search_once is True and (continuous is True or search_forever is not None):
        logger.error(
            "You cannot specify `--search-once` alongside `--continuous` or `--search-forever`"
        )
        sys.exit(1)

    if any(
        [
            len(notifications) > 0,
            search_forever is not None,
            notify_first_try is not None,
            polling_interval is not None,
            search_once is True,
        ]
    ):
        continuous = True
    return continuous, search_windows, days_of_the_week


def _get_provider_kwargs_from_cli(
    rec_area: Tuple[Union[str, int]],
    campground: Tuple[Union[str, int]],
    campsite: Tuple[Union[str, int]],
    start_date: str,
    end_date: str,
    weekends: bool,
    nights: int,
    provider: Optional[str],
    continuous: bool,
    polling_interval: Optional[str],
    notifications: Tuple[str],
    notify_first_try: Optional[str],
    search_forever: Optional[str],
    search_once: bool,
    yaml_config: Optional[str],
    offline_search: bool,
    offline_search_path: Optional[str],
    equipment: Tuple[Union[str, int]],
    equipment_id: Tuple[Union[str, int]],
    day: Optional[Tuple[str]],
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """
    Get Provider kwargs from CLI
    """
    notifications = make_list(notifications)
    continuous, search_windows, days_of_the_week = _validate_campsites(
        rec_area=rec_area,
        campground=campground,
        campsite=campsite,
        start_date=start_date,
        end_date=end_date,
        weekends=weekends,
        nights=nights,
        provider=provider,
        yaml_config=yaml_config,
        continuous=continuous,
        polling_interval=polling_interval,
        notifications=notifications,
        notify_first_try=notify_first_try,
        search_forever=search_forever,
        search_once=search_once,
        day=day,
    )
    if len(notifications) == 0:
        notifications = ["silent"]
    if polling_interval is None:
        polling_interval = SearchConfig.RECOMMENDED_POLLING_INTERVAL
    if notify_first_try is None:
        notify_first_try = False
    else:
        notify_first_try = True
    if search_forever is None:
        search_forever = False
    else:
        search_forever = True
    provider_kwargs = {
        "search_window": search_windows,
        "recreation_area": make_list(rec_area),
        "campgrounds": make_list(campground),
        "campsites": make_list(campsite),
        "weekends_only": weekends,
        "nights": int(nights),
        "offline_search": offline_search,
        "offline_search_path": offline_search_path,
        "equipment": equipment,
        "equipment_id": equipment_id,
        "days_of_the_week": days_of_the_week,
    }
    search_kwargs = {
        "log": True,
        "verbose": True,
        "continuous": continuous,
        "polling_interval": float(polling_interval),
        "notify_first_try": notify_first_try,
        "notification_provider": notifications,
        "search_forever": search_forever,
        "search_once": search_once,
    }
    return provider_kwargs, search_kwargs


@camply_command_line.command(cls=RichCommand)
@rec_area_argument
@campground_argument
@campsite_id_argument
@start_date_argument
@end_date_argument
@nights_argument
@weekends_argument
@day_of_the_week_argument
@notifications_argument
@continuous_argument
@search_forever_argument
@yaml_config_argument
@offline_search_argument
@offline_search_path_argument
@search_once_argument
@polling_interval_argument
@notify_first_try_argument
@equipment_argument
@equipment_id_argument
@provider_argument
@debug_option
@click.pass_obj
def campsites(
    context: CamplyContext,
    debug: bool,
    rec_area: Tuple[Union[str, int]],
    campground: Tuple[Union[str, int]],
    campsite: Tuple[Union[str, int]],
    start_date: str,
    end_date: str,
    weekends: bool,
    nights: int,
    provider: Optional[str],
    continuous: bool,
    polling_interval: Optional[str],
    notifications: Tuple[str],
    notify_first_try: Optional[str],
    search_forever: Optional[str],
    search_once: bool,
    yaml_config: Optional[str],
    offline_search: bool,
    offline_search_path: Optional[str],
    equipment: Tuple[Union[str, int]],
    equipment_id: Tuple[Union[str, int]],
    day: Optional[Tuple[str]],
) -> None:
    """
    Find Available Campsites with Custom Search Criteria

    Search for a campsite within camply. Campsites are returned based on the search criteria
    provided. Campsites contain properties like booking date, site type (tent, RV, cabin, etc),
    capacity, price, and a link to make the booking. Required parameters include
    `--start-date`, `--end-date`, `--rec-area` / `--campground`. Constant searching
    functionality can be enabled with  `--continuous` and notifications can be enabled using
    `--notifications`.
    """
    if context.debug is None:
        context.debug = debug
        _set_up_debug(debug=context.debug)
    if yaml_config is not None:
        provider, provider_kwargs, search_kwargs = yaml_utils.yaml_file_to_arguments(
            file_path=yaml_config
        )
        provider = _preferred_provider(context, provider)
    else:
        provider = _preferred_provider(context, provider)
        provider_kwargs, search_kwargs = _get_provider_kwargs_from_cli(
            rec_area=rec_area,
            campground=campground,
            campsite=campsite,
            start_date=start_date,
            end_date=end_date,
            weekends=weekends,
            nights=nights,
            provider=provider,
            continuous=continuous,
            polling_interval=polling_interval,
            notifications=notifications,
            notify_first_try=notify_first_try,
            search_forever=search_forever,
            search_once=search_once,
            offline_search=offline_search,
            offline_search_path=offline_search_path,
            equipment=equipment,
            equipment_id=equipment_id,
            day=day,
            yaml_config=yaml_config,
        )
    provider_class: Type[BaseCampingSearch] = CAMPSITE_SEARCH_PROVIDER[provider]
    camping_finder: BaseCampingSearch = provider_class(**provider_kwargs)
    camping_finder.get_matching_campsites(**search_kwargs)


@camply_command_line.command(cls=RichCommand)
@debug_option
@click.pass_obj
def providers(
    context: CamplyContext,
    debug: bool,
) -> None:
    """
    List the different camply providers
    """
    if context.debug is None:
        context.debug = debug
        _set_up_debug(debug=context.debug)
    logger.info(
        "camply currently supports %s providers:", len(CAMPSITE_SEARCH_PROVIDER.keys())
    )
    for provider_name, search_class in CAMPSITE_SEARCH_PROVIDER.items():
        logger.info(
            '    "%s":    %s',
            provider_name,
            search_class.__doc__.strip().splitlines()[0],
        )


test_notifications_kwargs = notification_kwargs.copy()
test_notifications_kwargs["help"] = test_notifications_kwargs["help"].replace(
    "Enables continuous searching. ", ""
)
test_notifications_kwargs.pop("default")
test_notifications_kwargs["required"] = True


@camply_command_line.command(cls=RichCommand)
@click.option("--notifications", **test_notifications_kwargs)
@debug_option
@click.pass_obj
def test_notifications(
    context: CamplyContext, debug: bool, notifications: Container[str]
) -> None:
    """
    Test your notification provider setup
    """
    if context.debug is None:
        context.debug = debug
        _set_up_debug(debug=context.debug)
    notification_providers = make_list(notifications)
    provider = MultiNotifierProvider(provider=notification_providers)
    logger.info("Testing your notification providers:")
    for sub_provider in provider.providers:
        logger.info('\t"%s"', sub_provider)
    provider.send_campsites(campsites=[example_campsite])


@camply_command_line.command(cls=RichCommand)
@rec_area_argument
@campground_argument
@provider_argument
@debug_option
@click.pass_obj
def list_campsites(
    context: CamplyContext,
    debug: bool,
    campground: Sequence[int],
    rec_area: Sequence[int],
    provider: str,
) -> None:
    """
    List campsite IDs for a given campground or recreation area
    """
    provider = _preferred_provider(context, provider)
    if context.debug is None:
        context.debug = debug
        _set_up_debug(debug=context.debug)
    if all(
        [
            len(rec_area) == 0,
            len(campground) == 0,
        ]
    ):
        logger.error(
            "You must provide a `--campground` or `--rec-area` to list campsites"
        )
        sys.exit(1)
    search_provider_class = CAMPSITE_SEARCH_PROVIDER[provider]
    if search_provider_class.list_campsites_supported is False:
        logger.error("That provider does not support listing campsites")
        sys.exit(1)
    camp_search = search_provider_class(
        search_window=SearchWindow(
            start_date=date.today() + timedelta(days=1),
            end_date=date.today() + timedelta(days=2),
        ),
        recreation_area=rec_area,
        campgrounds=campground,
        verbose=False,
    )
    logger.info("Searching for campsites to list")
    camp_search.list_campsite_units()


def cli():
    """
    Camply Command Line Utility Wrapper
    """
    try:
        camply_command_line()
    except KeyboardInterrupt:
        logger.debug("Handling Exit Request")
    finally:
        logger.camply("Exiting camply 👋")


if __name__ == "__main__":
    cli()
